import io
import csv
from decimal import InvalidOperation
from .rules import (
    SourceNotFound
)


ERROR_MSG_TEMPLATE = "Error executing:\n\tRule: {}\n{}:\n\t{}"


class CSVConverter:
    """Essentially a wrapper around a set of rules"""

    def __init__(self, rules):
        self.rules = rules

    def _write_to_outfile(self, data, outfile):
        """Internal method to write converted results to outfile.

        Args:
            data (str/dict): The data to be written.
            outfile (str): The file path to write the result to.

        """
        file = open(outfile, 'w')
        file.write(str(data))
        file.close()

    def _convert_to_csv(self, data):
        """Internal method to convert json data into csv data.

        Args:
            data (dict): The data to be converted.

        """
        csv_result = io.StringIO()
        field_names = []
        for rule in self.rules:
            field_names.append(rule.target)
        writer = csv.DictWriter(csv_result, field_names)
        writer.writeheader()
        for row in data:
            writer.writerow(row)
        return csv_result.getvalue()

    def convert(self, csv_file, to='json', outfile=None):
        """Example function with types documented in the docstring.

        Args:
            csv_file (str): The file path of the csv file to convert.
            to (str): What to return the output as, either `csv` or `json`. Default `json`.
            outfile (str): Optional - The file path to write the result to.

        Returns:
            str/dict: Depending on to `to parameter`
        """
        result = []
        source_file = open(csv_file)
        reader = csv.DictReader(source_file)

        for row in reader:
            row_result = {}
            for rule in self.rules:
                try:
                    k, v = rule.execute(row)
                    row_result[k] = v
                except SourceNotFound:
                    error_msg = ERROR_MSG_TEMPLATE.format(str(rule.as_dict()), "Unable to retrieve source data from", str(row))
                    print(error_msg)
                    row_result[rule.target] = ""
                except InvalidOperation:
                    error_msg = ERROR_MSG_TEMPLATE.format(str(rule.as_dict()), "Unable to convert data type for", str(row))
                    print(error_msg)
                    row_result[rule.target] = ""
                except ValueError:
                    error_msg = ERROR_MSG_TEMPLATE.format(str(rule.as_dict()), "Unable to convert data type for", str(row))
                    print(error_msg)
                    row_result[rule.target] = ""

            result.append(row_result)

        if to == 'csv':
            result = self._convert_to_csv(result)

        if outfile:
            self._write_to_outfile(result, outfile)

        source_file.close()
        return result



from decimal import Decimal
from datetime import datetime
from enum import Enum

import yaml


class OutputType(Enum):
    String = "String"
    Integer = "Integer"
    Decimal = "Decimal"
    Date = "Date"


class SourceNotFound(Exception):
    pass


class Rule:
    """
    Class representation of a Rule

    ...

    Attributes
    ----------
    source : str
        the key to use to pull from the data source
    target : str
        the key to use for the resulting data set
    output_type : OutputType
        the python data type you want the value to conform to
    """
    def __init__(self, source=None, target=None, output_type=OutputType.String, operations=None):
        self.source = source
        self.target = target
        self.output_type = output_type
        self.operations = operations

    def _value_as_output_type(self, value):
        if self.output_type == OutputType.Integer:
            return int(value)
        if self.output_type == OutputType.String:
            return str(value)
        if self.output_type == OutputType.Decimal:
            if type(value) is Decimal:
                return value
            if type(value) is str:
                value = value.replace(',', '')
            return Decimal(value)

    def execute(self, data):
        '''Implemented in subclasses'''
        pass # pragma: no cover

    def as_dict(self):
        '''Return a Dictionary representation of the rule'''
        result = {
            'target': self.target,
            'to_type': self.output_type.value,
            'source': self.source,
            'operations': self.operations
        }
        return result


class StaticRule(Rule):
    """
    Use Case: Static value for every row
    """
    def execute(self, data):
        """Returns the static value with self.target

        Args:
            data (dict): The data to perform to rule on.

        Returns:
            str, result: A k,v pair representing the target, and the retrieved/calculated value

        Raises:
            SourceNotFound: If the source cannot be found in the given data.
        """
        """Returns self.target, self.value as a key, value pair with self.value being cast to self.output_type"""
        return self.target, self._value_as_output_type(self.source)

    def as_dict(self):
        result = super().as_dict()
        result['type'] = 'static'
        return result


class SingleSourceRule(Rule):
    """
    Use Case: Extract value from column, place in new column
    """

    def execute(self, data):
        """Attempts to retrieve self.source from data.

        Args:
            data (dict): The data to perform to rule on.

        Returns:
            str, result: A k,v pair representing the target, and the retrieved/calculated value

        Raises:
            SourceNotFound: If the source cannot be found in the given data.
        """
        if self.source in data:
            value = data[self.source]
            if self.operations:
                for operation in self.operations:
                    value = eval(operation, {'s': value})
            return self.target, self._value_as_output_type(value)
        else:
            raise SourceNotFound

    def as_dict(self):
        result = super().as_dict()
        result['type'] = 'single'
        return result

class CalculationRule(Rule):
    """
    Use Case: Perform arithmetic calculations with values from columns
    """

    def execute(self, data):
        """Attempts to retrieve all of the values in self.source, then performs the calculations defined in self.operations.

        Args:
            data (dict): The data to perform to rule on.

        Returns:
            str, Decimal: A k,v pair representing the target, and the retrieved/calculated value

        Raises:
            SourceNotFound: If the any of the keys in self.source cannot be found in the given data.
        """
        values = []
        for key in self.source:
            if key in data:
                v = data[key]
                v = v.replace(',', '')
                v = Decimal(v)
                values.append(v)
            else:
                raise SourceNotFound
        value = eval(self.operations[0], {'s': values})
        return self.target, self._value_as_output_type(value)

    def as_dict(self):
        result = super().as_dict()
        result['type'] = 'multiple'
        return result


class DateRule(Rule):
    """
    Use Case: Gather date part values from multiple columns, generate date value
    """

    def execute(self, data):
        """Attempts to retrieve day, month, and year values from data,

        Args:
            data (dict): The data to perform to rule on.

        Returns:
            str, datetime: A k,v pair representing the target, and the retrieved/calculated value

        Raises:
            SourceNotFound: If the source cannot be found in the given data.
        """
        day_source = self.source['day']
        month_source = self.source['month']
        year_source = self.source['year']

        if day_source in data and month_source in data and year_source in data:
            day = int(data[day_source])
            month = int(data[month_source])
            year = int(data[year_source])
            return self.target, datetime(year, month, day)
        else:
            raise SourceNotFound

    def as_dict(self):
        result = super().as_dict()
        result['type'] = 'multiple'
        return result


def load_rules_from_yaml(file_name):
    """Generates a list of rules given a yaml configuration.

    Args:
        file_name (str): The file path of the file to use.

    Returns:
        list: A list of rules based on the given configuration.
    """
    file = open(file_name)
    definition = yaml.full_load(file)
    file.close()
    rules = []
    for rule in definition['rules']:
        target = rule['target']
        rule_type = rule['type']
        output_type = OutputType(rule['to_type'])
        source = rule['source']
        operations = rule['operations'] if 'operations' in rule else None

        if rule_type == 'multiple':
            if output_type == OutputType.Date:
                date_rule = DateRule(
                    source={
                        'day': source['day'],
                        'month': source['month'],
                        'year': source['year']
                    },
                    target=target,
                    output_type=output_type,
                    operations=operations
                )
                rules.append(date_rule)

            if output_type == OutputType.Decimal:
                calculation_rule = CalculationRule(
                    source=source,
                    target=target,
                    output_type=output_type,
                    operations=operations
                )
                rules.append(calculation_rule)

        elif rule_type == 'single':
            single_source_rule = SingleSourceRule(
                source=source,
                target=target,
                output_type=output_type,
                operations=operations
            )
            rules.append(single_source_rule)

        elif rule_type == 'static':
            static_rule = StaticRule(
                source=source,
                target=target,
                output_type=output_type,
                operations=operations
            )
            rules.append(static_rule)

        else:
            print('Unknown rule definition')
            print('Target: {}\n Type: {}\n Output Type: {}\n Source: {}'.format(target, rule_type, output_type, source))

    return rules


def export_rules_as_yaml(rules):
    """Converts a set of rules into a yaml configuration.

    Args:
        rules (list): The list of rules to convert.

    Returns:
        str: The yaml representation of the given rules.
    """
    rule_defs = []
    for rule in rules:
        rule_defs.append(rule.as_dict())
    return yaml.dump({'rules': rule_defs})



import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name='csv_etl',
    packages=find_packages(include=['csv_etl']),
    version='0.0.2',
    description='',
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/winslowdibona/csv_etl",
    author='Winslow DiBona',
    license='MIT',
    install_requires=['pyyaml'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests',
)
